# -*- coding: utf-8 -*-
#    Yum Exteder (yumex) - A graphic package management tool
#    Copyright (C) 2013 -2014 Tim Lauridsen < timlau<AT>fedoraproject<DOT>org >
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to
#    the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import glob
import logging
import os
import shutil

from gi.repository import GObject, Gtk
from yumex import const
from yumex.misc import CONFIG, _, format_number

from .views import RepoView

logger = logging.getLogger('yumex.gui.dialogs')


class AboutDialog(Gtk.AboutDialog):

    def __init__(self, parent):
        Gtk.AboutDialog.__init__(self)
        self.props.program_name = 'Yum Extender (dnf)'
        self.props.version = const.VERSION
        self.props.authors = ['Tim Lauridsen <timlau@fedoraproject.org>']
        self.props.license_type = Gtk.License.GPL_2_0
        self.props.copyright = '(C) 2021 Tim Lauridsen'
        self.props.website = 'https://www.yumex.dk'
        self.props.logo_icon_name = 'yumex-dnf'
        self.set_transient_for(parent)


class ProgressSplash:

    def __init__(self, base):
        self.base = base
        self.win = self.base.ui.get_object("win_working")
        pix = self.base.ui.get_object("work_pix")
        pix_file = f'{const.PIX_DIR}/progress.gif'
        pix.set_from_file(pix_file)
        self.label = self.base.ui.get_object("work_label")
        self.sublabel = self.base.ui.get_object("work_sublabel")
        self.win.set_transient_for(self.base)

    def show(self):
        self.win.show()

    def hide(self):
        self.win.hide()

    def set_label(self, text):
        self.label.set_text(text)

    def set_sublabel(self, text):
        self.sublabel.set_text(text)


class Preferences:

    VALUES = ['update_interval', 'refresh_interval', 'installonly_limit']
    FLAGS = ['autostart', 'clean_unused', 'newest_only',
             'headerbar', 'auto_select_updates', 'repo_saved', 'clean_instonly', 'use_dark', 'search_visible', 'show_splash'
             ]

    def __init__(self, base):
        self.base = base
        self.dialog = self.base.ui.get_object("preferences")
        self.dialog.set_transient_for(base)
        self.repo_view = RepoView()
        widget = self.base.ui.get_object('repo_sw')
        widget.add(self.repo_view)
        self.repo_box = self.base.ui.get_object("box_repos")
        # track when repo page is active in stack
        self.repo_box.connect("map", self.on_repo_page_active)
        self.repos = []

    def run(self):
        self.get_settings()
        self.dialog.show_all()
        rc = self.dialog.run()
        self.dialog.hide()
        need_reset = False
        if rc == Gtk.ResponseType.OK:
            need_reset = self.set_settings()
        return need_reset

    def on_repo_page_active(self, widget, *args):
        """ Callback for ::map event there is called when repo page is active"""
        if not self.repos:
            self._load_repositories()

    def _load_repositories(self):
        """ Lazy load repositories """
        # get the repositories
        self.base.set_working(True, splash=True)
        self.base.infobar.message(_('Fetching repository information'))
        self.repos = self.base.backend.get_repositories()
        self.base.infobar.hide()
        self.repo_view.populate(self.repos)
        self.base.set_working(False, splash=True)
        if CONFIG.conf.repo_saved:
            self.repo_view.select_by_keys(CONFIG.session.enabled_repos)

    def get_themes(self):
        # Get Themes
        pattern = os.path.normpath(os.path.join(const.THEME_DIR, '*.theme'))
        theme_files = glob.glob(pattern)
        theme_names = [os.path.basename(theme).split('.')[0]
                       for theme in theme_files]
        widget = self.base.ui.get_object('pref_theme')
        widget.remove_all()
        default = CONFIG.conf.theme.split(".")[0]
        i = 0
        ndx = 0
        for theme in sorted(theme_names):
            widget.append_text(theme)
            if theme == default:
                ndx = i
            i += 1
        widget.set_active(ndx)

    def get_settings(self):
        # set boolean states
        for option in Preferences.FLAGS:
            logger.debug("%s : %s ", option, getattr(CONFIG.conf, option))
            widget = self.base.ui.get_object('pref_' + option)
            widget.set_active(getattr(CONFIG.conf, option))
        # cleanup installonly handler
        widget = self.base.ui.get_object('pref_clean_instonly')
        widget.connect('notify::active', self.on_clean_instonly)
        # Set value states
        for name in Preferences.VALUES:
            widget = self.base.ui.get_object('pref_' + name)
            widget.set_value(getattr(CONFIG.conf, name))
        self.on_clean_instonly()
        # Get Themes
        self.get_themes()

    def on_clean_instonly(self, *args):
        """Handler for clean_instonly switch"""
        widget = self.base.ui.get_object('pref_clean_instonly')
        state = widget.get_active()
        postfix = 'installonly_limit'
        self._set_sensitive(postfix, state)

    def _set_sensitive(self, postfix, state):
        for prefix in ['pref_', 'label_']:
            id_ = prefix + postfix
            if state:
                self.base.ui.get_object(id_).set_sensitive(True)
            else:
                self.base.ui.get_object(id_).set_sensitive(False)

    def set_settings(self):
        changed = False
        need_reset = False
        # handle boolean options
        for option in Preferences.FLAGS:
            widget = self.base.ui.get_object('pref_' + option)
            state = widget.get_active()
            if state != getattr(CONFIG.conf, option):  # changed ??
                setattr(CONFIG.conf, option, state)
                changed = True
                self.handle_setting(option, state)
        # handle value options
        for name in Preferences.VALUES:
            widget = self.base.ui.get_object('pref_' + name)
            value = widget.get_value_as_int()
            if value != getattr(CONFIG.conf, name):  # changed ??
                setattr(CONFIG.conf, name, value)
                changed = True
        # handle repos, if the repositories has been loaded
        if self.repos:
            repo_before = CONFIG.session.enabled_repos
            repo_now = self.repo_view.get_selected()
            # repo selection changed
            if repo_now != repo_before:
                CONFIG.session.enabled_repos = repo_now     # set the new selection
                # we need to reset the gui
                need_reset = True
                if CONFIG.conf.repo_saved:
                    CONFIG.conf.repo_enabled = repo_now
                    changed = True
        # Themes
        widget = self.base.ui.get_object('pref_theme')
        default = CONFIG.conf.theme.split(".")[0]
        theme = widget.get_active_text()
        if theme != default:
            CONFIG.conf.theme = f'{theme}.theme'
            self.base.load_custom_styling()
            changed = True
        if changed:
            CONFIG.write()
        return need_reset

    def handle_setting(self, option, state):
        if option == 'autostart':
            if state:  # create an autostart .desktop for current user
                if not os.path.isdir(const.AUTOSTART_DIR):
                    logger.info("creating autostart directory : %s",
                                const.AUTOSTART_DIR)
                    os.makedirs(const.AUTOSTART_DIR, 0o700)
                shutil.copy(const.SYS_DESKTOP_FILE, const.USER_DESKTOP_FILE)
            else:  # remove the autostart file
                if os.path.exists(const.USER_DESKTOP_FILE):
                    os.unlink(const.USER_DESKTOP_FILE)


class ErrorDialog:

    def __init__(self, base):
        self.base = base
        self.dialog = self.base.ui.get_object("error_dialog")
        self.dialog.set_transient_for(base)
        self._buffer = self.base.ui.get_object('error_buffer')

    def show(self, txt):
        self._set_text(txt)
        self.dialog.show_all()
        rc = self.dialog.run()
        self.dialog.hide()
        self._buffer.set_text('')
        return rc == Gtk.ResponseType.CLOSE

    def _set_text(self, txt):
        self._buffer.set_text(txt)


class TransactionResult:

    def __init__(self, base):
        self.base = base
        self.dialog = self.base.ui.get_object("transaction-results")
        self.dialog.set_transient_for(base)
        self.view = self.base.ui.get_object("result_view")
        self.store = self.setup_view(self.view)

    def run(self):
        self.dialog.show_all()
        rc = self.dialog.run()
        self.dialog.hide()
        return rc == Gtk.ResponseType.OK

    def clear(self):
        self.store.clear()

    def setup_view(self, view):
        """
        Setup the TreeView
        @param view: the TreeView widget
        """
        model = Gtk.TreeStore(GObject.TYPE_STRING, GObject.TYPE_STRING,
                              GObject.TYPE_STRING, GObject.TYPE_STRING,
                              GObject.TYPE_STRING)
        view.set_model(model)
        self.create_text_column(_("Name"), view, 0, size=250)
        self.create_text_column(_("Arch"), view, 1)
        self.create_text_column(_("Ver"), view, 2)
        self.create_text_column(_("Repository"), view, 3, size=100)
        self.create_text_column(_("Size"), view, 4)
        return model

    def create_text_column(self, hdr, view, colno, size=None):
        """
        Create at TreeViewColumn
        @param hdr: column header text
        @param view: the TreeView widget
        @param colno: the TreeStore column containing data for the column
        @param size: the min column view (optional)
        """
        cell = Gtk.CellRendererText()  # Size Column
        column = Gtk.TreeViewColumn(hdr, cell, markup=colno)
        column.set_resizable(True)
        if size:
            column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            column.set_fixed_width(size)
        view.append_column(column)

    def populate(self, pkglist, dnl_size):
        """
        Populate the TreeView with data
        @param pkglist: list containing view data
        @param dnl_size:
        """
        model = self.store
        self.store.clear()
        total_size = 0
        for sub, lvl1 in pkglist:
            label = "<b>%s</b>" % const.TRANSACTION_RESULT_TYPES[sub]
            level1 = model.append(None, [label, "", "", "", ""])
            for pkgid, size, replaces in lvl1:
                (n, e, v, r, a, repo_id) = str(pkgid).split(',')
                level2 = model.append(
                    level1, [n, a, "%s.%s" % (v, r), repo_id,
                             format_number(size)])
                # packages there need to be downloaded
                if sub in ['install', 'update', 'install-deps',
                           'update-deps', 'obsoletes']:
                    total_size += size
                for r in replaces:
                    (n, e, v, r, a, repo_id) = str(r).split(',')
                    model.append(level2, [_("<b>replacing</b> {}").format(n),
                                          a, "%s.%s" % (v, r), repo_id,
                                          format_number(size)])
        self.base.ui.get_object("result_size").set_text(
            format_number(total_size))
        self.view.expand_all()


def show_information(window, msg, add_msg=None):
    dialog = Gtk.MessageDialog(
        flags=0, message_type=Gtk.MessageType.INFO,
        buttons=Gtk.ButtonsType.OK, text=msg)
    if add_msg:
        dialog.format_secondary_text(add_msg)
    if window:
        dialog.set_transient_for(window)
    dialog.run()
    dialog.destroy()


def yes_no_dialog(window, msg, add_msg=None):
    dialog = Gtk.MessageDialog(
        flags=0, message_type=Gtk.MessageType.INFO,
        buttons=Gtk.ButtonsType.YES_NO, text=msg)
    if add_msg:
        dialog.format_secondary_text(add_msg)
    if window:
        dialog.set_transient_for(window)
    rc = dialog.run()
    dialog.destroy()
    return rc == Gtk.ResponseType.YES


def ask_for_gpg_import(window, values):
    (pkg_id, userid, hexkeyid, keyurl, timestamp) = values
    pkg_name = pkg_id.split(',')[0]
    msg = (_(' Do you want to import this GPG key\n'
             ' needed to verify the %s package?\n\n'
             ' Key        : 0x%s:\n'
             ' Userid     : "%s"\n'
             ' From       : %s') %
           (pkg_name, hexkeyid, userid,
            keyurl.replace("file://", "")))

    dialog = Gtk.MessageDialog(
        window, 0, Gtk.MessageType.QUESTION, Gtk.ButtonsType.YES_NO, msg)
    rc = dialog.run()
    dialog.destroy()
    return rc == Gtk.ResponseType.YES
